package com.yonge.netty.dto;

import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.stream.Collectors;

import javax.sound.sampled.AudioFormat;

import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.yonge.audio.analysis.AudioFloatConverter;
import com.yonge.audio.analysis.Signals;
import com.yonge.audio.analysis.detector.PercussionRhythmDetector;
import com.yonge.audio.analysis.detector.YINPitchDetector;
import com.yonge.netty.dto.NoteAnalysis.NoteErrorType;
import com.yonge.netty.entity.MusicXmlBasicInfo;
import com.yonge.netty.entity.MusicXmlNote;
import com.yonge.netty.entity.MusicXmlSection;
import com.yonge.netty.server.processor.WaveformWriter;

import be.tarsos.dsp.pitch.FastYin;

/**
 * 用户通道上下文
 */
public class UserChannelContext {

	private final static Logger LOGGER = LoggerFactory.getLogger(UserChannelContext.class);

	// 打击乐
	// private final static List<Integer> percussionList = Arrays.asList(23, 113, 121);

	private final static int MIN_FREQUECY = 43;

	private final static int MAX_FREQUECY = 2000;
	
	private AudioFormat audioFormat = new AudioFormat(44100, 16, 1, true, false);
	
	private int bufferSize = 1024;
	
	private int frameSize = 128;

	private FastYin detector;
	
	private AudioFloatConverter converter;

	private String user;

	private double standardFrequecy = 442;

	private float offsetMS;

	private float micDelayMS;

	private double dynamicOffset;

	private String platform;

	private Long recordId;

	private Integer subjectId;

	private String evaluationCriteria;

	private float beatDuration;

	private boolean isProcessedDynamicOffset;

	// 曲目与musicxml对应关系
	private ConcurrentHashMap<Integer, MusicXmlBasicInfo> songMusicXmlMap = new ConcurrentHashMap<Integer, MusicXmlBasicInfo>();

	private WaveformWriter waveFileProcessor;

	private NoteAnalysis processingNote = new NoteAnalysis(0, 0, -1);

	private AtomicInteger evaluatingSectionIndex = new AtomicInteger(0);

	private List<NoteAnalysis> doneNoteAnalysisList = new ArrayList<NoteAnalysis>();

	private List<SectionAnalysis> doneSectionAnalysisList = new ArrayList<SectionAnalysis>();

	private List<ChunkAnalysis> totalChunkAnalysisList = new ArrayList<ChunkAnalysis>();
	
	private List<Float> floatSamples = new ArrayList<Float>();
	
	private byte[] channelBufferBytes = new byte[0];

	private double playTime;

	private HardLevelEnum hardLevel = HardLevelEnum.BEGINNER;

	private boolean handlerSwitch;

	private NotePlayResult queryNoteFrequency(MusicXmlNote xmlNote, double playFrequency) {

		NotePlayResult result = new NotePlayResult();

		boolean status = false;
		double migrationRate = 0;

		if (Math.round(xmlNote.getFrequency()) == Math.round(playFrequency)) {
			status = true;
			migrationRate = 0;
		} else {
			NoteFrequencyRange noteFrequencyRange = new NoteFrequencyRange(standardFrequecy, xmlNote.getFrequency());

			if (noteFrequencyRange.getMinFrequency() > playFrequency || playFrequency > noteFrequencyRange.getMaxFrequency()) {
				status = false;
			} else {

				status = true;

				if (Math.round(playFrequency) < Math.round(xmlNote.getFrequency())) {
					double min = Math.abs(xmlNote.getFrequency() - noteFrequencyRange.getMinFrequency()) / 2;
					migrationRate = Math.abs(playFrequency - xmlNote.getFrequency()) / min;
				} else {
					double max = Math.abs(xmlNote.getFrequency() - noteFrequencyRange.getMaxFrequency()) / 2;
					migrationRate = Math.abs(playFrequency - xmlNote.getFrequency()) / max;
				}
			}
		}

		result.setStatus(status);
		result.setMigrationRate(migrationRate);

		return result;
	}

	public void init(MusicXmlBasicInfo musicXmlBasicInfo) {
		this.platform = musicXmlBasicInfo.getPlatform();
		this.subjectId = musicXmlBasicInfo.getSubjectId();
		this.beatDuration = musicXmlBasicInfo.getBeatLength();
		this.hardLevel = HardLevelEnum.valueOf(musicXmlBasicInfo.getHeardLevel());
		this.evaluationCriteria = musicXmlBasicInfo.getEvaluationCriteria();
	}

	public void setUser(String user) {
		this.user = user;
	}

	public Long getRecordId() {
		return recordId;
	}

	public void setRecordId(Long recordId) {
		this.recordId = recordId;
	}

	public boolean getHandlerSwitch() {
		return handlerSwitch;
	}

	public void setHandlerSwitch(boolean handlerSwitch) {
		this.handlerSwitch = handlerSwitch;
	}

	public float getOffsetMS() {
		return offsetMS;
	}

	public void setOffsetMS(float offsetMS) {
		this.offsetMS = offsetMS;
	}

	public float getMicDelayMS() {
		return micDelayMS;
	}

	public void setMicDelayMS(float micDelayMS) {
		this.micDelayMS = micDelayMS;
	}

	public float getBeatDuration() {
		return beatDuration;
	}

	public void setBeatDuration(float beatDuration) {
		this.beatDuration = beatDuration;
	}

	public HardLevelEnum getHardLevel() {
		return hardLevel;
	}

	public ConcurrentHashMap<Integer, MusicXmlBasicInfo> getSongMusicXmlMap() {
		return songMusicXmlMap;
	}

	public WaveformWriter getWaveFileProcessor() {
		return waveFileProcessor;
	}

	public void setWaveFileProcessor(WaveformWriter waveFileProcessor) {
		this.waveFileProcessor = waveFileProcessor;
	}

	public NoteAnalysis getProcessingNote() {
		return processingNote;
	}

	public void setProcessingNote(NoteAnalysis processingNote) {
		this.processingNote = processingNote;
	}

	public List<SectionAnalysis> getDoneSectionAnalysisList() {
		return doneSectionAnalysisList;
	}

	public List<NoteAnalysis> getDoneNoteAnalysisList() {
		return doneNoteAnalysisList;
	}

	public void resetUserInfo() {
		waveFileProcessor = null;
		processingNote = new NoteAnalysis(0, 0, -1);
		evaluatingSectionIndex = new AtomicInteger(0);
		channelBufferBytes = new byte[0];
		doneNoteAnalysisList = new ArrayList<NoteAnalysis>();
		doneSectionAnalysisList = new ArrayList<SectionAnalysis>();
		totalChunkAnalysisList = new ArrayList<ChunkAnalysis>();
		floatSamples = new ArrayList<Float>();
		recordId = null;
		playTime = 0;
		isProcessedDynamicOffset = false;
		offsetMS = 0;
		dynamicOffset = 0;
		handlerSwitch = false;
	}

	public MusicXmlBasicInfo getMusicXmlBasicInfo(Integer songId) {
		MusicXmlBasicInfo musicXmlBasicInfo = null;
		if (songId == null) {
			musicXmlBasicInfo = songMusicXmlMap.values().stream().findFirst().get();
		} else {
			musicXmlBasicInfo = songMusicXmlMap.get(songId);
		}
		return musicXmlBasicInfo;
	}

	public MusicXmlSection getCurrentMusicSection(Integer songId, int sectionIndex) {
		MusicXmlBasicInfo musicXmlBasicInfo = getMusicXmlBasicInfo(songId);
		return musicXmlBasicInfo.getMusicXmlSectionMap().get(sectionIndex);
	}

	public MusicXmlNote getCurrentMusicNote(Integer songId, Integer noteIndex) {
		if (songMusicXmlMap.size() == 0) {
			return null;
		}
		if (noteIndex == null) {
			noteIndex = processingNote.getMusicalNotesIndex();
		}
		final int index = noteIndex;
		MusicXmlBasicInfo musicXmlBasicInfo = getMusicXmlBasicInfo(songId);

		int totalNoteIndex = getTotalMusicNoteIndex(null);
		if (musicXmlBasicInfo != null && index <= totalNoteIndex) {
			return musicXmlBasicInfo.getMusicXmlInfos().stream().filter(t -> t.getMusicalNotesIndex() == index).findFirst().get();
		}

		return null;
	}
	
	public int getTotalMusicNoteIndex(Integer songId) {
		if (songMusicXmlMap.size() == 0) {
			return -1;
		}
		MusicXmlBasicInfo musicXmlBasicInfo = getMusicXmlBasicInfo(songId);

		if (musicXmlBasicInfo != null) {
			return musicXmlBasicInfo.getMusicXmlInfos().stream().map(t -> t.getMusicalNotesIndex()).distinct().max(Integer::compareTo).get();
		}

		return -1;
	}

	public List<MusicXmlNote> getCurrentMusicSection(Integer songId, Integer sectionIndex) {
		if (songMusicXmlMap.size() == 0) {
			return null;
		}
		if (sectionIndex == null) {
			sectionIndex = processingNote.getSectionIndex();
		}
		final int index = sectionIndex;
		MusicXmlBasicInfo musicXmlBasicInfo = getMusicXmlBasicInfo(songId);

		if (musicXmlBasicInfo != null) {
			return musicXmlBasicInfo.getMusicXmlInfos().stream().filter(t -> t.getMusicalNotesIndex() == index)
					.sorted(Comparator.comparing(MusicXmlNote::getMusicalNotesIndex)).collect(Collectors.toList());
		}

		return null;
	}

	public int getTotalMusicSectionSize(Integer songId) {
		if (songMusicXmlMap.size() == 0) {
			return -1;
		}
		MusicXmlBasicInfo musicXmlBasicInfo = getMusicXmlBasicInfo(songId);

		if (musicXmlBasicInfo != null) {
			return (int) musicXmlBasicInfo.getMusicXmlInfos().stream().map(t -> t.getMeasureIndex()).distinct().count();
		}

		return -1;
	}

	public int getMusicSectionIndex(Integer songId, int musicXmlNoteIndex) {
		if (songMusicXmlMap.size() == 0) {
			return -1;
		}

		if (getTotalMusicNoteIndex(null) < musicXmlNoteIndex) {
			return -1;
		}

		MusicXmlBasicInfo musicXmlBasicInfo = getMusicXmlBasicInfo(songId);

		if (musicXmlBasicInfo != null) {
			return musicXmlBasicInfo.getMusicXmlInfos().stream().filter(t -> t.getMusicalNotesIndex() == musicXmlNoteIndex).findFirst().get().getMeasureIndex();
		}

		return -1;
	}

	public byte[] getChannelBufferBytes() {
		return channelBufferBytes;
	}

	public void setChannelBufferBytes(byte[] channelBufferBytes) {
		this.channelBufferBytes = channelBufferBytes;
	}

	public AtomicInteger getEvaluatingSectionIndex() {
		return evaluatingSectionIndex;
	}

	public int getBufferSize() {
		return bufferSize;
	}

	public void handle(byte[] datas) {
		
		//applyLowPassFilter(datas, 3000, audioFormat.getSampleRate());
		//applyMovingAverageFilter(datas, 5);
		//datas = applyNoiseGate(datas, 1000);
		
		if(converter == null) {
			converter = AudioFloatConverter.getConverter(audioFormat);
		}
		
		float[] samples = new float[getBufferSize()];

		converter.toFloatArray(datas, samples);
		
		for(float f : samples) {
			floatSamples.add(f);
		}
		
		if (detector == null) {
			detector = new FastYin(audioFormat.getSampleRate(), getBufferSize());
		}

		// YINPitchDetector frequencyDetector = new YINPitchDetector(samples.length , audioFormat.getSampleRate());
		// int playFrequency = (int) frequencyDetector.getFrequency(samples);

		float playFrequency = -1;
		if (StringUtils.equalsIgnoreCase(evaluationCriteria, EvaluationCriteriaEnum.FREQUENCY.getCode())) {
			playFrequency = detector.getPitch(samples).getPitch();
		}

		double splDb = 0;
		float[] energyEnvelop = Signals.energyEnvelope(samples, frameSize);
		float amplitude = Signals.norm(samples);
		float power = Signals.power(samples);
		
		int decibels = 0;
		
		double durationTime = 1000 * (samples.length * 2) / audioFormat.getSampleRate() / (audioFormat.getSampleSizeInBits() / 8);

		playTime += durationTime;
		
		// 获取当前音符信息
		MusicXmlNote musicXmlNote = getCurrentMusicNote(null, null);

		if (musicXmlNote == null) {
			return;
		}

		// 取出当前处理中的音符信息
		NoteAnalysis noteAnalysis = getProcessingNote();
		if (noteAnalysis == null || noteAnalysis.getDurationTime() == 0) {
			noteAnalysis = new NoteAnalysis(musicXmlNote.getMeasureRenderIndex(), musicXmlNote.getMusicalNotesIndex(), musicXmlNote.getMeasureIndex(),
					(int) musicXmlNote.getFrequency(), musicXmlNote.getDuration());
		}

		evaluatingSectionIndex.set(noteAnalysis.getSectionIndex());

		if (noteAnalysis.getMusicalNotesIndex() >= 0 && noteAnalysis.getMusicalNotesIndex() <= getTotalMusicNoteIndex(null)) {

			ChunkAnalysis chunkAnalysis = new ChunkAnalysis(playTime - durationTime, playTime, playFrequency, splDb, energyEnvelop, amplitude);

			if (totalChunkAnalysisList.size() > 0) {
				if (totalChunkAnalysisList.get(totalChunkAnalysisList.size() - 1).getAmplitude() + 2 < chunkAnalysis.getAmplitude()) {
					chunkAnalysis.setPeak(true);// 只针对打击乐
				}
			}
			
			//处理泛音
			float basicFrequency = -1;
			if(totalChunkAnalysisList.size() > 0) {
				basicFrequency = totalChunkAnalysisList.get(totalChunkAnalysisList.size() - 1).getFrequency();
			}
			chunkAnalysis.setFrequency(handleHarmonic(basicFrequency, playFrequency));
			chunkAnalysis.setFrequency(handleHarmonic(musicXmlNote.getFrequency(), playFrequency));
			
			totalChunkAnalysisList.add(chunkAnalysis);

			LOGGER.debug("user:{}  isProcessedDynamicOffset:{}  dynamicOffset:{}  Frequency:{}  splDb:{}  amplitude:{} power:{} decibels:{} endtime:{}", user,
					isProcessedDynamicOffset, dynamicOffset, chunkAnalysis.getFrequency(), splDb, amplitude, power, decibels, playTime);

			if (playTime >= (musicXmlNote.getDuration() + musicXmlNote.getTimeStamp() + getOffsetMS() + beatDuration)) {

				musicXmlNote.setTimeStamp(musicXmlNote.getTimeStamp() + getOffsetMS() + beatDuration);
				
				if (musicXmlNote.getDontEvaluating()) {
					noteAnalysis.setIgnore(true);
				}
				
				List<ChunkAnalysis> nodeChunkAnalysisList = extract(musicXmlNote, noteAnalysis);
				
				// 判断节奏(音符持续时间内有不间断的音高，就节奏正确)
				if (!StringUtils.equalsIgnoreCase(evaluationCriteria, EvaluationCriteriaEnum.FREQUENCY.getCode())) {
					noteAnalysis.setTempoStatus(computeTempoWithAmplitude(musicXmlNote, noteAnalysis, nodeChunkAnalysisList));
					noteAnalysis.setPlayFrequency(-1);
				} else {
					noteAnalysis.setTempoStatus(computeTempoWithFrequency(musicXmlNote, noteAnalysis, nodeChunkAnalysisList));
					noteAnalysis.setPlayFrequency(computeFrequency(musicXmlNote, noteAnalysis, nodeChunkAnalysisList));
				}

				evaluateForNote(musicXmlNote, noteAnalysis, nodeChunkAnalysisList);// 对当前音符评分

				LOGGER.debug("当前小节[{}] 音符下标[{}] 预计频率：{} 实际频率：{} 节奏：{}", musicXmlNote.getMeasureIndex(), noteAnalysis.getMusicalNotesIndex(), musicXmlNote.getFrequency(),
						noteAnalysis.getPlayFrequency(), noteAnalysis.getTempoStatus());

				doneNoteAnalysisList.add(noteAnalysis);

				// 准备处理下一个音符
				int nextNoteIndex = musicXmlNote.getMusicalNotesIndex() + 1;
				float nextNoteFrequence = -1;
				double standDuration = 0;
				int measureRenderIndex = 0;
				MusicXmlNote nextMusicXmlNote = getCurrentMusicNote(null, nextNoteIndex);
				if (nextMusicXmlNote != null) {
					nextNoteFrequence = nextMusicXmlNote.getFrequency();
					standDuration = nextMusicXmlNote.getDuration();
					measureRenderIndex = nextMusicXmlNote.getMeasureRenderIndex();
				}

				NoteAnalysis nextNoteAnalysis = new NoteAnalysis(measureRenderIndex, nextNoteIndex, getMusicSectionIndex(null, nextNoteIndex),
						(int) nextNoteFrequence, standDuration);

				noteAnalysis = nextNoteAnalysis;

			}

			setProcessingNote(noteAnalysis);
		}

	}

	public int evaluateForSection(int sectionIndex, int subjectId) {

		int score = -1;
		if (doneSectionAnalysisList.size() >= getTotalMusicSectionSize(null)) {
			return score;
		}

		// 取出当前小节的所有音符
		List<NoteAnalysis> noteAnalysisList = doneNoteAnalysisList.stream().filter(t -> t.getSectionIndex() == sectionIndex).collect(Collectors.toList());

		long ignoreSize = noteAnalysisList.stream().filter(t -> t.isIgnore()).count();

		SectionAnalysis sectionAnalysis = new SectionAnalysis();
		sectionAnalysis.setIndex(sectionIndex);
		sectionAnalysis.setNoteNum(noteAnalysisList.size());
		sectionAnalysis.setIsIngore(ignoreSize == noteAnalysisList.size());

		// 判断是否需要评分
		MusicXmlSection musicXmlSection = getCurrentMusicSection(null, sectionIndex);
		if (noteAnalysisList.size() == musicXmlSection.getNoteNum()) {
			sectionAnalysis.setMeasureRenderIndex(noteAnalysisList.stream().findFirst().get().getMeasureRenderIndex());
			// 取出需要评测的音符
			List<NoteAnalysis> noteList = noteAnalysisList.stream().filter(t -> t.isIgnore() == false).collect(Collectors.toList());

			if (noteList != null && noteList.size() > 0) {
				score = noteList.stream().mapToInt(t -> t.getScore()).sum() / noteList.size();
			}
			sectionAnalysis.setDurationTime(noteAnalysisList.stream().mapToDouble(t -> t.getDurationTime()).sum());
			sectionAnalysis.setScore(score);

			LOGGER.debug("小节评分：{}", sectionAnalysis);
			doneSectionAnalysisList.add(sectionAnalysis);
		}

		return score;
	}

	public Map<String, Integer> evaluateForMusic() {

		Map<String, Integer> result = new HashMap<String, Integer>();

		result.put("playTime", (int) doneNoteAnalysisList.stream().mapToDouble(t -> t.getDurationTime()).sum());
		result.put("recordId", recordId.intValue());

		// 取出需要评测的音符
		List<NoteAnalysis> noteAnalysisList = doneNoteAnalysisList.stream().filter(t -> t.isIgnore() == false).collect(Collectors.toList());

		if (noteAnalysisList != null && noteAnalysisList.size() > 0) {
			int intonationScore = 0;
			int tempoScore = 0;
			int integrityScore = 0;
			int socre = 0;

			for (NoteAnalysis note : noteAnalysisList) {
				intonationScore += note.getIntonationScore();
				tempoScore += note.getTempoScore();
				integrityScore += note.getIntegrityScore();
				socre += note.getScore();
			}

			tempoScore = tempoScore / noteAnalysisList.size();
			intonationScore = intonationScore / noteAnalysisList.size();
			integrityScore = integrityScore / noteAnalysisList.size();

			result.put("cadence", tempoScore);
			result.put("intonation", intonationScore);
			result.put("integrity", integrityScore);

			int score = socre / noteAnalysisList.size();

			// 平均得分
			if (getMusicXmlBasicInfo(null).getSubjectId() == 23 || getMusicXmlBasicInfo(null).getSubjectId() == 113) {
				score = tempoScore;
			}
			result.put("score", score);
		} else {
			result.put("cadence", 0);
			result.put("intonation", 0);
			result.put("integrity", 0);
			result.put("score", 0);
		}
		return result;
	}

	public void evaluateForNote(MusicXmlNote musicXmlNote, NoteAnalysis noteAnalysis, List<ChunkAnalysis> chunkAnalysisList) {

		double durationTime = chunkAnalysisList.get(chunkAnalysisList.size() - 1).getEndTime() - chunkAnalysisList.get(0).getStartTime();

		double playDurationTime = 0;

		if (StringUtils.equalsIgnoreCase(evaluationCriteria, EvaluationCriteriaEnum.FREQUENCY.getCode())) {

			NotePlayResult notePlayResult = queryNoteFrequency(musicXmlNote, noteAnalysis.getPlayFrequency());

			if (noteAnalysis.getFrequency() == -1) {// 休止符

				playDurationTime = chunkAnalysisList.stream().filter(t -> t.getFrequency() <= MIN_FREQUECY).mapToDouble(t -> t.getDurationTime()).sum();

				if (noteAnalysis.getTempoStatus() == 0) {
					noteAnalysis.setMusicalErrorType(NoteErrorType.CADENCE_WRONG);
				} else if (playDurationTime * 100 / durationTime < hardLevel.getIntegrityRange()) {
					noteAnalysis.setMusicalErrorType(NoteErrorType.INTEGRITY_WRONG);
				} else if (notePlayResult.getStatus() == false) {
					noteAnalysis.setMusicalErrorType(NoteErrorType.INTONATION_WRONG);
				} else {
					noteAnalysis.setMusicalErrorType(NoteErrorType.RIGHT);
				}
			} else {
				playDurationTime = chunkAnalysisList.stream().filter(t -> t.getFrequency() > MIN_FREQUECY && t.getFrequency() < MAX_FREQUECY)
						.mapToDouble(t -> t.getDurationTime()).sum();

				if (playDurationTime * 100 / durationTime < hardLevel.getNotPlayRange()) {
					noteAnalysis.setMusicalErrorType(NoteErrorType.NOT_PLAY);
					LOGGER.debug("未演奏：{}", playDurationTime * 100 / durationTime);
				} else if (playDurationTime * 100 / durationTime < hardLevel.getIntegrityRange()) {
					noteAnalysis.setMusicalErrorType(NoteErrorType.INTEGRITY_WRONG);
					LOGGER.debug("完整度不足：{}", playDurationTime * 100 / durationTime);
				} else if (noteAnalysis.getTempoStatus() == 0) {
					noteAnalysis.setMusicalErrorType(NoteErrorType.CADENCE_WRONG);
				} else if (notePlayResult.getStatus() == false) {
					noteAnalysis.setMusicalErrorType(NoteErrorType.INTONATION_WRONG);
				} else {
					noteAnalysis.setMusicalErrorType(NoteErrorType.RIGHT);
				}
			}

		}

		// 计算音分
		int tempoScore = 0;
		int integrityScore = 0;
		int intonationScore = 100 - new BigDecimal(Math
				.abs(YINPitchDetector.hertzToAbsoluteCent(noteAnalysis.getPlayFrequency()) - YINPitchDetector.hertzToAbsoluteCent(noteAnalysis.getFrequency())))
						.multiply(new BigDecimal(20)).divide(new BigDecimal(17), BigDecimal.ROUND_UP).setScale(0, BigDecimal.ROUND_UP).intValue();
		if (intonationScore < 0) {
			intonationScore = 0;
		} else if (intonationScore > 100) {
			intonationScore = 100;
		}

		if (noteAnalysis.getMusicalErrorType() == NoteErrorType.NOT_PLAY) {
			intonationScore = 0;
		} else {

			if (noteAnalysis.getTempoStatus() != 0) {
				tempoScore = 100;
				noteAnalysis.setTempoScore(tempoScore);
			}

			integrityScore = (int) (playDurationTime * 100 * 100 / hardLevel.getIntegrityRange() / durationTime);
			if (integrityScore > 100) {
				integrityScore = 100;
			}
			noteAnalysis.setIntegrityScore(integrityScore);
		}
		noteAnalysis.setIntonationScore(intonationScore);
		if (!StringUtils.equalsIgnoreCase(evaluationCriteria, EvaluationCriteriaEnum.FREQUENCY.getCode())) {
			noteAnalysis.setScore(tempoScore);
		} else {
			noteAnalysis.setScore(
					new BigDecimal(intonationScore + tempoScore + integrityScore).divide(new BigDecimal(3), 2).setScale(0, BigDecimal.ROUND_UP).intValue());
		}
	}

	private float computeFrequency(MusicXmlNote musicXmlNote, NoteAnalysis noteAnalysis, List<ChunkAnalysis> chunkAnalysisList) {

		if (chunkAnalysisList == null || chunkAnalysisList.size() == 0 || musicXmlNote.getDontEvaluating()) {
			return -1;
		}

		List<ChunkAnalysis> chunkList = null;
		
		/**
		if(noteAnalysis.getRhythmStartTime() > 0) {
			chunkList = chunkAnalysisList.stream().filter(t -> t.getStartTime() >= noteAnalysis.getRhythmStartTime()).collect(Collectors.toList());
		} else {
			chunkList = new ArrayList<ChunkAnalysis>(chunkAnalysisList);
		}*/
		
		chunkAnalysisList = totalChunkAnalysisList.stream()
				.filter(t -> Double.doubleToLongBits(t.getStartTime()) >= Double.doubleToLongBits(noteAnalysis.getRhythmStartTime())
						&& Double.doubleToLongBits(t.getEndTime()) <= Double.doubleToLongBits(noteAnalysis.getRhythmStartTime() + musicXmlNote.getDuration()))
				.collect(Collectors.toList());

		reduceNoise(chunkAnalysisList, EvaluationCriteriaEnum.FREQUENCY);
		
		int range = hardLevel.getTempoEffectiveRange(musicXmlNote.getDenominator(), musicXmlNote.getDuration());
		
		chunkList = chunkAnalysisList.subList(0, chunkAnalysisList.size() * (100 - range) / 100);
		
		LOGGER.debug("根据节奏点[{}]，取当前音符有效的信号范围[ {} - {} ]计算评率", noteAnalysis.getRhythmStartTime(), chunkList.get(0).getStartTime(), chunkList.get(chunkList.size() - 1).getEndTime());

		List<Float> chunkFrequencyList = chunkList.stream().map(t -> t.getFrequency())
				.filter(t -> t.doubleValue() > MIN_FREQUECY && t.doubleValue() < MAX_FREQUECY).collect(Collectors.toList());

		if (chunkFrequencyList.size() == 0) {
			return -1;
		}

		float frequency = chunkFrequencyList.stream().reduce(0f, (a, b) -> a + b) / chunkFrequencyList.size();
		
		return frequency;
	}

	/**
	 * 时值范围内有且只有一个音，且不能间断，且在合理范围内需开始演奏
	 * 与上一个音相同时，2个音之间需要间断
	 * @param musicXmlNote
	 * @return
	 */
	private int computeTempoWithFrequency(MusicXmlNote musicXmlNote, NoteAnalysis noteAnalysis, List<ChunkAnalysis> chunkAnalysisList) {
		
		if (chunkAnalysisList == null || chunkAnalysisList.size() == 0 || musicXmlNote.getDontEvaluating()) {
			noteAnalysis.setMusicalErrorType(NoteErrorType.NOT_PLAY);
			return -1;
		}
		
		LOGGER.debug("---------------------TEMPO----------------------");
		
		ChunkAnalysis firstChunkAnalysis = chunkAnalysisList.get(0);
		
		noteAnalysis.setRhythmStartTime(musicXmlNote.getTimeStamp() + dynamicOffset);//默认值
		
		Map<Integer, Double> frequencyRhythmMap = queryRhythmsByFrequency(musicXmlNote, noteAnalysis, chunkAnalysisList);

		if(musicXmlNote.getFrequency() < MIN_FREQUECY) {
			if(frequencyRhythmMap.size() == 0) {
				noteAnalysis.setMusicalErrorType(NoteErrorType.RIGHT);
				return 99;
			} else {
				noteAnalysis.setMusicalErrorType(NoteErrorType.CADENCE_WRONG);
				LOGGER.debug("根据能量包络检测到[{}]个断点，分别在[{}]", frequencyRhythmMap.size(), frequencyRhythmMap.values().stream().map(value -> value).map(Object::toString).collect(Collectors.joining(",")));
				return 0;
			}
		}
		
		if(frequencyRhythmMap.size() > 1) {
			noteAnalysis.setMusicalErrorType(NoteErrorType.CADENCE_WRONG);
			LOGGER.debug("根据音高检测到[{}]个断点，分别在[{}]", frequencyRhythmMap.size(), frequencyRhythmMap.values().stream().map(value -> value).map(Object::toString).collect(Collectors.joining(",")));
			return 0;
		}
		
		double firstBeatTime = 0;
		
		if(frequencyRhythmMap.size() == 1) {
			// 判断进入时间点
			for(Entry<Integer, Double> entry : frequencyRhythmMap.entrySet()) {
				firstBeatTime = entry.getValue();
			}
			
			noteAnalysis.setRhythmStartTime(firstBeatTime);
			
			if ((firstBeatTime - firstChunkAnalysis.getStartTime()) * 100 / musicXmlNote.getDuration() > hardLevel.getTempoEffectiveRange(musicXmlNote.getDenominator(),
					musicXmlNote.getDuration()) * 2) {
				noteAnalysis.setMusicalErrorType(NoteErrorType.CADENCE_WRONG);
				LOGGER.debug("节奏错误原因：进入时间点[{}]太晚", firstBeatTime);
				return 0;
			}
			noteAnalysis.setMusicalErrorType(NoteErrorType.RIGHT);
			LOGGER.debug("找到节奏点StartTime:{}", firstBeatTime);
		} else {
			noteAnalysis.setMusicalErrorType(NoteErrorType.NOT_PLAY);
			LOGGER.debug("没有找到节奏点");
			return 0;
		}

		/**
		//根据能量包络再进行分析节奏
		Map<Integer, Double> energyRrhythmMap = queryRhythmsByEnergyEnvelope(musicXmlNote, noteAnalysis, chunkAnalysisList);
		
		if (energyRrhythmMap.size() > 1) {
			LOGGER.debug("根据能量包络检测到[{}]个节奏点，分别在[{}]", energyRrhythmMap.size(), energyRrhythmMap.values().stream().map(value -> value + firstChunkAnalysis.getStartTime()).map(Object::toString).collect(Collectors.joining(",")));
			return 0;
		}
		
		if(energyRrhythmMap.size() == 1) {
			for(Entry<Integer, Double> entry : energyRrhythmMap.entrySet()) {
				firstBeatTime = entry.getValue();
			}
			
			firstBeatTime = firstChunkAnalysis.getStartTime() + firstBeatTime;

			// 判断进入时间点
			if ((firstBeatTime - firstChunkAnalysis.getStartTime()) * 100 / musicXmlNote.getDuration() > hardLevel.getTempoEffectiveRange(musicXmlNote.getDenominator(),
					musicXmlNote.getDuration()) * 2) {
				LOGGER.debug("节奏错误原因：进入时间点[{}]太晚", firstBeatTime);
				return 0;
			}
		}
		
		if(energyRrhythmMap.size() == 0 && frequencyRhythmMap.size() == 0) {
			LOGGER.debug("节奏错误原因：没有找到节奏点");
			return 0;
		}
		*/
		
		if(isProcessedDynamicOffset == false) {

			double startTime = musicXmlNote.getTimeStamp() + dynamicOffset;
			dynamicOffset = firstBeatTime - startTime;
			isProcessedDynamicOffset = true;
		}
		
		return 99;
	}
	

	private int computeTempoWithAmplitude(MusicXmlNote musicXmlNote, NoteAnalysis noteAnalysis, List<ChunkAnalysis> chunkAnalysisList) {

		if (chunkAnalysisList == null || chunkAnalysisList.size() == 0 || musicXmlNote.getDontEvaluating()) {
			noteAnalysis.setMusicalErrorType(NoteErrorType.NOT_PLAY);
			return 0;
		}
		
		//Map<Integer, Double> rhythmMap = queryRhythmsByEnergyEnvelope(musicXmlNote, noteAnalysis, chunkAnalysisList);
		
		Map<Integer, Double> rhythmMap = queryRhythmsByAmplitude(musicXmlNote, noteAnalysis, chunkAnalysisList);
		
		ChunkAnalysis firstChunkAnalysis = chunkAnalysisList.get(0);
		
		if(musicXmlNote.getFrequency() < MIN_FREQUECY) {
			if(rhythmMap.size() == 0) {
				noteAnalysis.setMusicalErrorType(NoteErrorType.RIGHT);
				return 3;
			} else {
				noteAnalysis.setMusicalErrorType(NoteErrorType.CADENCE_WRONG);
				LOGGER.debug("根据能量包络检测到[{}]个断点，分别在[{}]", rhythmMap.size(), rhythmMap.values().stream().map(value -> value + firstChunkAnalysis.getStartTime()).map(Object::toString).collect(Collectors.joining(",")));
				return 0;
			}
		}
		
		if(rhythmMap.size() != 1) {
			LOGGER.debug("根据能量包络检测到[{}]个断点，分别在[{}]", rhythmMap.size(), rhythmMap.values().stream().map(value -> value + firstChunkAnalysis.getStartTime()).map(Object::toString).collect(Collectors.joining(",")));
			
			if(rhythmMap.size() > 1) {
				noteAnalysis.setMusicalErrorType(NoteErrorType.CADENCE_WRONG);
			}else {
				noteAnalysis.setMusicalErrorType(NoteErrorType.NOT_PLAY);
			}
			return 0;
		}
		
		// 判断进入时间点
		double firstBeatTime = 0;
		for(Entry<Integer, Double> entry : rhythmMap.entrySet()) {
			firstBeatTime = entry.getValue();
		}
		
		firstBeatTime = firstChunkAnalysis.getStartTime() + firstBeatTime;
		
		if ((firstBeatTime - firstChunkAnalysis.getStartTime()) * 100 / musicXmlNote.getDuration() > hardLevel.getTempoEffectiveRange(musicXmlNote.getDenominator(),
				musicXmlNote.getDuration()) * 2) {
			LOGGER.debug("节奏错误原因：进入时间点[{}]太晚", firstBeatTime);
			
			noteAnalysis.setMusicalErrorType(NoteErrorType.CADENCE_WRONG);
			return 0;
		}
		
		LOGGER.debug("找到节奏点StartTime:{}", firstBeatTime);
		
		noteAnalysis.setMusicalErrorType(NoteErrorType.RIGHT);
		
		if(isProcessedDynamicOffset == false) {

			double startTime = musicXmlNote.getTimeStamp() + dynamicOffset;
			dynamicOffset = firstBeatTime - startTime;
			isProcessedDynamicOffset = true;
		}
		
		return 3;
	}

	private double queryFirstNoteStartTime(List<ChunkAnalysis> chunkAnalysisList, MusicXmlNote musicXmlNote, NoteAnalysis noteAnalysis) {

		if (chunkAnalysisList == null || chunkAnalysisList.size() == 0) {
			LOGGER.debug("找不到数据,StartTime:{}", musicXmlNote.getTimeStamp() + dynamicOffset);
			return musicXmlNote.getTimeStamp() + dynamicOffset;
		}
		
		if (!StringUtils.equalsIgnoreCase(evaluationCriteria, EvaluationCriteriaEnum.FREQUENCY.getCode())) {
			return Math.max(chunkAnalysisList.get(0).getStartTime(), 0);
		}

		double onsetStartTime = 0;
		double preNoteAvgPower = 0;
		
		Map<Integer, ChunkAnalysis> peakMap = new HashMap<Integer, ChunkAnalysis>();
		
		/**
		ChunkAnalysis firstChunkAnalysis = chunkAnalysisList.get(0);
		
		Optional<ChunkAnalysis> chunkAnalysisOptional = totalChunkAnalysisList.stream()
				.filter(t -> Double.doubleToLongBits(t.getEndTime()) <= Double.doubleToLongBits(firstChunkAnalysis.getStartTime()))
				.reduce((first, second) -> second);

		ChunkAnalysis lastChunkAnalysis = null;
		if (chunkAnalysisOptional.isPresent()) {
			lastChunkAnalysis = chunkAnalysisOptional.get();
		}

		if (lastChunkAnalysis == null) {
			lastChunkAnalysis = new ChunkAnalysis(0, 0, -1, -40, 0, 0);
		}
		
		chunkAnalysisList.add(0, lastChunkAnalysis);
		**/
		
		//获取上一个音符
		Optional<NoteAnalysis> preNoteAnalysisOptinal = doneNoteAnalysisList.stream().filter(t -> t.getIndex() == musicXmlNote.getMusicalNotesIndex() - 1).findFirst();
		
		if (StringUtils.equalsIgnoreCase(evaluationCriteria, EvaluationCriteriaEnum.FREQUENCY.getCode())) {
			//有音高，声压才有效
			float preNoteAvgFrequency = -1;
			
			if(preNoteAnalysisOptinal.isPresent()) {
				preNoteAvgPower = preNoteAnalysisOptinal.get().getDBSPL();
				preNoteAvgFrequency = preNoteAnalysisOptinal.get().getPlayFrequency();
			} else {
				preNoteAvgPower = -40;
			}
			
			LOGGER.debug("上一个音符的平均音高[{}] 声压[{}]", preNoteAvgFrequency, preNoteAvgPower);
			
			ChunkAnalysis chunkAnalysis = null;
			
			//既要找到起始点，又要判断节奏是否正确
			//若是休止符
			if(musicXmlNote.getFrequency() < MIN_FREQUECY) {
				for(int i = 0; i < chunkAnalysisList.size(); i++) {
					chunkAnalysis = chunkAnalysisList.get(i);
					
					if(chunkAnalysis.getFrequency() < MIN_FREQUECY) {
						onsetStartTime = chunkAnalysis.getStartTime();
						noteAnalysis.setTempoStatus(1);
						LOGGER.debug("范围内查询到音高信号,EndTime:{} ChunkIndex:{}", onsetStartTime, i);
						return onsetStartTime;
					}
				}
			}
			
			//若不是休止符
			if(musicXmlNote.getFrequency() >= MIN_FREQUECY) {
				for(int i = 0; i < chunkAnalysisList.size(); i++) {
					chunkAnalysis = chunkAnalysisList.get(i);
					
					if(chunkAnalysis.getFrequency() >= MIN_FREQUECY) {
						if (preNoteAvgFrequency < MIN_FREQUECY || !isSamePitch(preNoteAvgFrequency, chunkAnalysis.getFrequency())) {
							onsetStartTime = chunkAnalysis.getStartTime();
							noteAnalysis.setTempoStatus(1);
							LOGGER.debug("范围内查询到音高信号,EndTime:{} ChunkIndex:{}", onsetStartTime, i);
							return onsetStartTime;
						}
					}
				}
			}
			
			peakMap = detectPeaks(chunkAnalysisList, preNoteAvgPower + 1);
			
		} else {

			if(preNoteAnalysisOptinal.isPresent()) {
				preNoteAvgPower = preNoteAnalysisOptinal.get().getDBSPL();
			} else {
				preNoteAvgPower = -40;
			}
			
			peakMap = detectPeaks(chunkAnalysisList, preNoteAvgPower);
		}
		
		if (peakMap.size() > 0) {
			Integer firstPeakIndex = peakMap.keySet().stream().max(Integer::compare).get();
			onsetStartTime = chunkAnalysisList.get(firstPeakIndex).getStartTime();
			LOGGER.debug("范围内查询到声压信号,StartTime:{}", onsetStartTime);
			noteAnalysis.setTempoStatus(2);
			return onsetStartTime;
		}

		if (onsetStartTime == 0) {
			LOGGER.debug("范围内未查询到信号,StartTime:{}", musicXmlNote.getTimeStamp() + dynamicOffset);
		}

		// return chunkAnalysisList.get(chunkAnalysisList.size() - 1).getEndTime();
		return Math.max(musicXmlNote.getTimeStamp() + dynamicOffset, onsetStartTime);
	}
	
	private void applyLowPassFilter(byte[] buffer, int cutoffFrequency, float sampleRate) {
        int sampleSize = 2; // 16-bit audio
        int length = buffer.length / sampleSize;
        double rc = 1.0 / (cutoffFrequency * 2 * Math.PI);
        double dt = 1.0 / sampleRate;
        double alpha = dt / (rc + dt);
        short[] samples = new short[length];

        // 将字节数组转换为短整型数组
        for (int i = 0; i < length; i++) {
            samples[i] = (short) ((buffer[2 * i + 1] << 8) | (buffer[2 * i] & 0xff));
        }

        // 应用低通滤波器
        for (int i = 1; i < length; i++) {
            samples[i] = (short) (samples[i - 1] + alpha * (samples[i] - samples[i - 1]));
        }

        // 将处理后的短整型数组转换回字节数组
        for (int i = 0; i < length; i++) {
            buffer[2 * i] = (byte) (samples[i] & 0xff);
            buffer[2 * i + 1] = (byte) ((samples[i] >> 8) & 0xff);
        }
    }
	
	private static byte[] applyNoiseGate(byte[] buffer, int threshold) {
        byte[] filteredBuffer = new byte[buffer.length];
        for (int i = 0; i < buffer.length; i++) {
            if (Math.abs(buffer[i]) < threshold) {
                filteredBuffer[i] = 0; // 设置为静音
            } else {
                filteredBuffer[i] = buffer[i];
            }
        }
        return filteredBuffer;
    }
	
	// 移动平均滤波器
    private void applyMovingAverageFilter(byte[] buffer, int windowSize) {
        int sampleSize = 2; // 16-bit audio
        int length = buffer.length / sampleSize;
        short[] samples = new short[length];
        
        // 将字节数组转换为短整型数组
        for (int i = 0; i < length; i++) {
            samples[i] = (short) ((buffer[2 * i + 1] << 8) | (buffer[2 * i] & 0xff));
        }

        short[] filteredSamples = new short[length];

        // 应用移动平均滤波器
        for (int i = 0; i < length; i++) {
            int sum = 0;
            int count = 0;
            for (int j = i - windowSize / 2; j <= i + windowSize / 2; j++) {
                if (j >= 0 && j < length) {
                    sum += samples[j];
                    count++;
                }
            }
            filteredSamples[i] = (short) (sum / count);
        }

        // 将处理后的短整型数组转换回字节数组
        for (int i = 0; i < length; i++) {
            buffer[2 * i] = (byte) (filteredSamples[i] & 0xff);
            buffer[2 * i + 1] = (byte) ((filteredSamples[i] >> 8) & 0xff);
        }
    }
	
	

	
	private void reduceNoise(List<ChunkAnalysis> chunkAnalysisList, EvaluationCriteriaEnum criteria) {

		ChunkAnalysis chunkAnalysis = null;

		for (int i = 1; i < chunkAnalysisList.size(); i++) {
			if (i < chunkAnalysisList.size() - 1) {
				chunkAnalysis = chunkAnalysisList.get(i);

				if (EvaluationCriteriaEnum.AMPLITUDE == criteria) {
					if (chunkAnalysisList.get(i - 1).getAmplitude() == 0 && chunkAnalysisList.get(i + 1).getAmplitude() == 0
							&& chunkAnalysis.getAmplitude() > 0) {

						chunkAnalysis.setAmplitude(0);
						// chunkAnalysisList.set(i, chunkAnalysis);
					}
				} else if (EvaluationCriteriaEnum.FREQUENCY == criteria) {
					if (chunkAnalysisList.get(i - 1).getFrequency() == -1 && chunkAnalysisList.get(i + 1).getFrequency() == -1
							&& chunkAnalysis.getFrequency() > 0) {

						chunkAnalysis.setFrequency(-1);
						// chunkAnalysisList.set(i, chunkAnalysis);
					}
				}
			}
		}

	}

	private boolean isSamePitch(float basicFrequency, float frequency) {
		if (new BigDecimal(basicFrequency + "").equals(new BigDecimal(frequency + ""))) {
			return true;
		}
		
		return (basicFrequency != -1 && frequency != -1) && Math.abs(YINPitchDetector.getDeviationCent(basicFrequency, frequency)) < 50;
	}

	private Map<Integer, ChunkAnalysis> detectPeaks(List<ChunkAnalysis> signalList, double threshold) {
		Map<Integer, ChunkAnalysis> peaks = new HashMap<Integer, ChunkAnalysis>();
		
		LOGGER.debug("平均声压值：[{}}", threshold);
		
		/**
		int waveCrestNums = 0,waveTroughNums = 0;
		for (int i = 1; i < signalList.size() - 1; i++) {
			if (signalList.get(i).getSplDb() > signalList.get(i - 1).getSplDb() && signalList.get(i).getSplDb() >= signalList.get(i + 1).getSplDb()
					&& signalList.get(i).getSplDb() > threshold + 2) {
				waveCrestNums++;
				waveTroughNums = 0;
				if (waveCrestNums > 1) {
					peaks.put(i, signalList.get(i));
					LOGGER.debug("范围内查询到异常音波信号,StartTime:{} CurrentSplDb:{} Threshold:{}", signalList.get(i).getStartTime(), signalList.get(i).getSplDb(), threshold);
				}
			} else if (signalList.get(i).getSplDb() < signalList.get(i - 1).getSplDb() && signalList.get(i).getSplDb() <= signalList.get(i + 1).getSplDb()
					&& signalList.get(i).getSplDb() + 2 < threshold) {
				waveTroughNums++;
				waveCrestNums = 0;
				if (waveTroughNums >= 1) {
					peaks.put(i, signalList.get(i));
					LOGGER.debug("范围内查询到异常音波信号,StartTime:{} CurrentSplDb:{} Threshold:{}", signalList.get(i).getStartTime(), signalList.get(i).getSplDb(), threshold);
				}
			} else {
				waveCrestNums = 0;
				waveTroughNums = 0;
			}
		}*/
		
		int continueSignalThreshold = 2;
		int continueNums = 0;

		if (StringUtils.equalsIgnoreCase(evaluationCriteria, EvaluationCriteriaEnum.FREQUENCY.getCode())) {
			for (int i = 1; i < signalList.size(); i++) {
				//如果连续2个下降超过平均值
				if(threshold - signalList.get(i).getSplDb() < 0 && signalList.get(i).getSplDb() > signalList.get(i - 1).getSplDb()) {
					++continueNums;
				} else {
					if(continueNums >= continueSignalThreshold) {
						peaks.put(i, signalList.get(i));
						LOGGER.debug("【过程】范围内查询到声压信号,StartTime:{} CurrentSplDb:{} Threshold:{}" , signalList.get(i).getStartTime(), signalList.get(i - 1).getSplDb(), threshold);
					}
					continueNums = 0;
				}
			}
		} else {
			int waveCrestNums = 0;
			for (int i = 1; i < signalList.size() - 1; i++) {
				if (signalList.get(i).getSplDb() > signalList.get(i - 1).getSplDb() && signalList.get(i).getSplDb() >= signalList.get(i + 1).getSplDb()
						&& signalList.get(i).getSplDb() > threshold) {
					waveCrestNums++;
					if (waveCrestNums > 0) {
						peaks.put(i, signalList.get(i));
						LOGGER.debug("范围内查询到异常音波信号,StartTime:{} CurrentSplDb:{} Threshold:{}", signalList.get(i).getStartTime(), signalList.get(i).getSplDb(), threshold);
					}
				} else {
					waveCrestNums = 0;
				}
			}
		}
		
		return peaks;
	}
	
	
	private Map<Integer,Double> queryRhythmsByEnergyEnvelope(MusicXmlNote musicXmlNote, NoteAnalysis noteAnalysis, List<ChunkAnalysis> datas){
		
		LOGGER.debug("------------利用能量包络寻找节奏点[{} - {}]------------", noteAnalysis.getStartTime(), noteAnalysis.getEndTime());
		
		
		int start = (int) (noteAnalysis.getStartTime() / ((1 * 1000) / audioFormat.getSampleRate()));
		int end = (int) (noteAnalysis.getEndTime() * audioFormat.getSampleRate() / 1 / 1000);
		
		//取信号
		float[] fSamples = new float[end - start];
		int j = 0;
		for(int i = start; i < end; i++) {
			fSamples[j] = floatSamples.get(i - 1);
			j++;
		}
		
		float[] energyEnvelope = Signals.energyEnvelope(fSamples, frameSize);
		
		/**
		float[] energyEnvelope = new float[datas.size() * getBufferSize() / frameSize];
		
		for(int i = 0; i < datas.size(); i++) {
			for(int j = 0; j < datas.get(i).getEnergyEnvelop().length; j++) {
				energyEnvelope[i * datas.get(i).getEnergyEnvelop().length + j] = datas.get(i).getEnergyEnvelop()[j];
			}
		}
		*/
		PercussionRhythmDetector percussionRhythmDetector = new PercussionRhythmDetector(energyEnvelope, audioFormat.getSampleRate(), frameSize);
		
		Map<Integer, Double> rhythMap = percussionRhythmDetector.detect();
		
		return rhythMap;
	}
	
	private Map<Integer,Double> queryRhythmsByAmplitude(MusicXmlNote musicXmlNote, NoteAnalysis noteAnalysis, List<ChunkAnalysis> datas){
		
		Map<Integer, Double> rhythMap = new HashMap<Integer, Double>();

		//reduceNoise(datas, EvaluationCriteriaEnum.AMPLITUDE);
		
		float amplitudeThreshold = 0.5f;
		int beatContinueNum = 0;
		int intervalTime = 50;
		ChunkAnalysis chunkAnalysis = null;
		double rhythmTime = -1;
		int peakIndex = 0;
		int continueNumThreshold = 0;
		boolean isContinue = false;
		
		for (int i = 0; i < datas.size(); i++) {
			chunkAnalysis = datas.get(i);

			if (chunkAnalysis.getAmplitude() >= amplitudeThreshold) {
				beatContinueNum++;
				if (beatContinueNum == 1) {
					rhythmTime = i * bufferSize * 1000 / audioFormat.getSampleRate();
				}
				
				if (beatContinueNum > continueNumThreshold && isContinue == false) {
					if (rhythMap.size() == 0 || rhythmTime - rhythMap.get(peakIndex) > intervalTime) {
						peakIndex++;
						rhythMap.put(peakIndex, rhythmTime);
						isContinue =true;
					}
				}

			} else {
				beatContinueNum = 0;
				isContinue = false;
			}
		}
		
		return rhythMap;
	}
	
	
	private Map<Integer, Double> queryRhythmsByFrequency(MusicXmlNote musicXmlNote, NoteAnalysis noteAnalysis, List<ChunkAnalysis> datas){
		LOGGER.debug("------------利用频率寻找节奏点------------");

		/**
		//获取上一个音符
		Optional<NoteAnalysis> preNoteAnalysisOptinal = doneNoteAnalysisList.stream().filter(t -> t.getIndex() == musicXmlNote.getMusicalNotesIndex() - 1).findFirst();
		
		float preNoteAvgFrequency = -1;
		if(preNoteAnalysisOptinal.isPresent()) {
			preNoteAvgFrequency = preNoteAnalysisOptinal.get().getPlayFrequency();
		}
		
		LOGGER.debug("上一个音符的平均音高[{}]", preNoteAvgFrequency);
		**/
		
		//获取上一个信号
		ChunkAnalysis firstChunkAnalysis = datas.get(0);
		
		Optional<ChunkAnalysis> chunkAnalysisOptional = totalChunkAnalysisList.stream()
				.filter(t -> Double.doubleToLongBits(t.getEndTime()) <= Double.doubleToLongBits(firstChunkAnalysis.getStartTime()))
				.reduce((first, second) -> second);

		ChunkAnalysis lastChunkAnalysis = null;
		if (chunkAnalysisOptional.isPresent()) {
			lastChunkAnalysis = chunkAnalysisOptional.get();
		}

		if (lastChunkAnalysis == null) {
			lastChunkAnalysis = new ChunkAnalysis(0, 0, -1, 0, new float[0], 0);
		}
		
		float preNoteAvgFrequency = lastChunkAnalysis.getFrequency();
		
		
		ChunkAnalysis chunkAnalysis = null;
		Map<Integer, Double> rhythMap = new HashMap<Integer, Double>();

		int intervalTime = 150;
		double lastestRhythmTime = Math.negateExact(intervalTime);

		int silenceContinueNum = 0;
		int beatContinueNum = 0;
		double rhythmTime = -1;
		int peakIndex = 0;
		int continueNumThreshold = 0;
		
		for (int i = 0; i < datas.size(); i++) {

			chunkAnalysis = datas.get(i);

			// 若不是休止符
			if (chunkAnalysis.getFrequency() < MIN_FREQUECY) {

				silenceContinueNum++;
				beatContinueNum = 0;

				if (silenceContinueNum > continueNumThreshold) {
					preNoteAvgFrequency = chunkAnalysis.getFrequency();
				}
			} else {

				silenceContinueNum = 0;

				if (preNoteAvgFrequency < MIN_FREQUECY || !isSamePitch(preNoteAvgFrequency, chunkAnalysis.getFrequency()) || musicXmlNote.getFrequency() < MIN_FREQUECY) {

					if (beatContinueNum == 0) {
						rhythmTime = chunkAnalysis.getStartTime();
					}

					beatContinueNum++;
					if (beatContinueNum > continueNumThreshold) {
						if (chunkAnalysis.getStartTime() - lastestRhythmTime > intervalTime) {

							lastestRhythmTime = rhythmTime;

							if (peakIndex == 0 || lastestRhythmTime - rhythMap.get(peakIndex) > intervalTime) {
								peakIndex++;
								rhythMap.put(peakIndex, lastestRhythmTime);
								LOGGER.debug("范围内查询到音高信号,preNoteFrequency:{} peakIndex:{} EndTime:{}", preNoteAvgFrequency, peakIndex, lastestRhythmTime);
							}

							preNoteAvgFrequency = chunkAnalysis.getFrequency();
							beatContinueNum = 0;
						}
					}
				} else {
					beatContinueNum = 0;
				}

			}
		}
			
			
		
		return rhythMap;
	}
	
	
	private List<ChunkAnalysis> extract(MusicXmlNote musicXmlNote, NoteAnalysis noteAnalysis) {
		
		LOGGER.debug("---------------------Extract Data----------------------");
		
		int range = hardLevel.getTempoEffectiveRange(musicXmlNote.getDenominator(), musicXmlNote.getDuration());
		
		double floatingRange = musicXmlNote.getDuration() * range / 100;

		double startTime = musicXmlNote.getTimeStamp() + dynamicOffset - floatingRange;
		
		double endTime = startTime + musicXmlNote.getDuration();
		
		/**
		//下一个音符信息
		MusicXmlNote nextMusicXmlNote = getCurrentMusicNote(null, musicXmlNote.getMusicalNotesIndex() + 1);
		if(nextMusicXmlNote != null) {
			endTime = endTime - nextMusicXmlNote.getDuration() * hardLevel.getTempoEffectiveRange(nextMusicXmlNote.getDenominator(), nextMusicXmlNote.getDuration()) / 100;
		}**/
		
		double finalEndTime = endTime;
		
		LOGGER.debug("当前音符有效信号时值[{}]偏移[{}]后的范围[ {} - {} ]", musicXmlNote.getDuration(), floatingRange, startTime, finalEndTime);
		
		List<ChunkAnalysis> chunkAnalysisList = totalChunkAnalysisList.stream()
				.filter(t -> Double.doubleToLongBits(t.getEndTime()) >= Double.doubleToLongBits(startTime)
						&& Double.doubleToLongBits(t.getStartTime()) <= Double.doubleToLongBits(finalEndTime))
				.collect(Collectors.toList());

		// 根据完整度取部分有效信号
		int elementSize = chunkAnalysisList.size() * (100 - range) / 100;
		
		List<ChunkAnalysis> datas = chunkAnalysisList.subList(0, elementSize);
		
		noteAnalysis.setStartTime(datas.get(0).getStartTime());
		noteAnalysis.setEndTime(datas.get(elementSize - 1).getEndTime());
		
		LOGGER.debug("根据用户的评测难度[{}]，当前音符有效的信号时值的取值范围[ {} - {} ]", range, datas.get(0).getStartTime(), datas.get(elementSize - 1).getEndTime());

		return datas;
	}
	
	private float handleHarmonic2(float basicFrequency, float frequency) {
		
		if (basicFrequency > frequency) {
			return frequency;
		}
		
		float threshold = 0.02f;
		// 处理泛音
		int roundedRatio = Math.round(frequency / basicFrequency);
		float ratio = frequency / basicFrequency;
		if (roundedRatio >= 2 && Math.abs(ratio - roundedRatio) <= threshold) {
			return frequency / ratio;
		}
		
		return frequency;
	}
	
	private float handleHarmonic(float basicFrequency, float frequency) {
		
		if(frequency < MIN_FREQUECY || basicFrequency < MIN_FREQUECY) {
			return frequency;
		}
		
		int roundedRatio = 0;
		float targetFrequency = frequency;
		
		if(basicFrequency > frequency) {
			roundedRatio = Math.round(basicFrequency / frequency);
			targetFrequency = frequency * roundedRatio;
		}else {
			roundedRatio = Math.round(frequency / basicFrequency);
			targetFrequency = frequency / roundedRatio;
		}
		
		if(isSamePitch(basicFrequency, targetFrequency)) {
			return targetFrequency;
		}
		
		return frequency;
	}
	
}
